﻿//#define DEBUG
//#define DEBUG_LINE
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SimpleJSON;
using UnityEngine;

namespace JointCorrect
{
    class JointCorrect : MVRScript
    {
        private Atom person;
        
        private List< MorphVariant > allMorphs = new List< MorphVariant >();
        private HashSet< string > allVariants = new HashSet< string >();
        private StorableStringSet enabledVariants;
        private Dictionary< string, JSONStorableBool > variantUIs;

        private MorphVariant[] 
            lFoot_x_p065, lFoot_x_n040,
            rFoot_x_p065, rFoot_x_n040,
            lShin_x_p085, lShin_x_p140,
            rShin_x_p085, rShin_x_p140,
            lThigh_x_p035, lThigh_x_n055, lThigh_x_n115, lThigh_y_p075, lThigh_y_n075, lThigh_z_p085, lThigh_z_n015,
            rThigh_x_p035, rThigh_x_n055, rThigh_x_n115, rThigh_y_n075, rThigh_y_p075, rThigh_z_n085, rThigh_z_p015,
            lForearm_y_n100, lForearm_y_n130,
            rForearm_y_p100, rForearm_y_p130,
            lShldr_y_n095, lShldr_y_p040, lShldr_z_n075, lShldr_z_n060, lShldr_z_p035,
            rShldr_y_p095, rShldr_y_n040, rShldr_z_p075, rShldr_z_p060, rShldr_z_n035,
            lCollar_x_n025, lCollar_x_p015, 
            rCollar_x_n025, rCollar_x_p015,
            lCollar_y_n026, lCollar_y_p017, 
            rCollar_y_n017, rCollar_y_p026,
            lCollar_z_n015, lCollar_z_p050, 
            rCollar_z_n050, rCollar_z_p015,
            pelvis_x_p030, pelvis_x_n015, 
            abdomen_x_n020, abdomen_x_p030, 
            abdomen2_x_n020, abdomen2_x_p030, 
            chest_x_p020,
            neck_x_n030,
            head_x_n045,
            
            cThighsXn115, cThighsZp180, cThighsZn030;

        private List< Action > drivers = new List< Action >();
        private List< string > NotFound = new List< string >();

        private StringBuilder rLog = new StringBuilder();
        private StringBuilder mLog = new StringBuilder();
        private JSONStorableString morphOut;
        private JSONStorableString jointOut;
        private JSONStorableBool locked;
        private JSONStorableBool builtin;

        private bool debug = true;
        private bool debugLine = false;
        private LineRenderer line;
        
        public override void Init()
        {
            base.Init();

            try
            { 
                person = GetContainingAtom();

                allVariants = new HashSet<string>();
                allVariants.Add( "" );
                allVariants.Add( "gens" );
                enabledVariants = new StorableStringSet( "enabledVariants", VariantsChanged );
                enabledVariants.Add( "" );
                enabledVariants.Add( "gens" );
            
                InitMorphs();
                RegisterString( enabledVariants );
            }
            catch( Exception ex )
            {
                SuperController.LogError( $"JointCorrect did not Init.\n{ex}" );
            }
        }

        public void InitMorphs()
        {
            try
            {
                locked = new JSONStorableBool( "Lock", false );
                builtin = new JSONStorableBool( "Built-in base", true );

                #if DEBUG
                { 
                    morphOut = new JSONStorableString( "morphout", "" );
                    var dbugField = CreateTextField( morphOut );
                    dbugField.height *= 4;

                    jointOut = new JSONStorableString( "jointout", "" );
                    var joutField = CreateTextField( jointOut, true );
                    joutField.height *= 4;

                    var lockedField = CreateToggle( locked, true );
                }
                #endif

                RegisterBool( builtin );
                CreateToggle( builtin, true );

                // upper body joints
                
                var lHand = GetJoint( "lHand" );
                var lForearm = GetJoint( "lForeArm" );
                var lShldr = GetJoint( "lShldr" );
                var lCollar = GetJoint( "lCollar" );
                var chest = GetJoint( "chest" );
                var rCollar = GetJoint( "rCollar" );
                var rShldr = GetJoint( "rShldr" );
                var rForearm = GetJoint( "rForeArm" );

                // trunk joints

                var hip = GetJoint( "hip" );
                var abdomen = GetJoint( "abdomen" );
                var abdomen2 = GetJoint( "abdomen2" );
                var neck = GetJoint( "neck" );
                var head = GetJoint( "head" );

                // lower body joints
                
                var lFoot = GetJoint( "lFoot" );
                var lShin = GetJoint( "lShin" );
                var lThigh = GetJoint( "lThigh" );
                var pelvis = GetJoint( "pelvis" );
                var rThigh = GetJoint( "rThigh" );
                var rShin = GetJoint( "rShin" );
                var rFoot = GetJoint( "rFoot" );

                #region Collar
                var lCollarBone = lCollar.GetComponent< DAZBone >();
                lCollar_x_n025 = GetMorph( "LCollarX-025" );
                lCollar_x_p015 = GetMorph( "LCollarX+015" );
                lCollar_y_n026 = GetMorph( "LCollarY-026" );
                lCollar_y_p017 = GetMorph( "LCollarY+017" );
                lCollar_z_n015 = GetMorph( "LCollarZ-015" );
                lCollar_z_p050 = GetMorph( "LCollarZ+050" );
                drivers.Add( () => { 
                    var reference = chest.rotation * lCollarBone.startingLocalRotation;

                    Vector3 rUp = (reference * Vector3.up).normalized,
                            rLeft = (reference * Vector3.left).normalized,
                            rForward = (reference * Vector3.forward).normalized;

                    var dazZRotation = AngleOffAroundAxis( lCollar.up, rUp, rForward, true );
                    PrintRotation( dazZRotation, "lCollar Z" );
                    var unZ = Quaternion.AngleAxis( dazZRotation, rForward );

                    //drawAxes( lCollar.rotation * unZ, lShldr.position );
                    SetMorph( lCollar_z_n015, dazZRotation, 0, -15 );
                    SetMorph( lCollar_z_p050, dazZRotation, 0, 50 );

                    var dazYRotation = AngleOffAroundAxis( unZ * (-lCollar.right), rLeft, rUp, true );
                    PrintRotation( dazYRotation, "lCollar Y" );
                    var unY = Quaternion.AngleAxis( dazYRotation, rUp );

                    SetMorph( lCollar_y_n026, dazYRotation, 0, -26 );
                    SetMorph( lCollar_y_p017, dazYRotation, 0, 17 );

                    var dazXRotation = AngleOffAroundAxis( unY * unZ * lCollar.up, rUp, rLeft, true );
                    PrintRotation( dazXRotation, "lCollar X" );

                    SetMorph( lCollar_x_n025, dazXRotation, 0, -25 );
                    SetMorph( lCollar_x_p015, dazXRotation, 0, 15 );
                    
                });

                var rCollarBone = rCollar.GetComponent< DAZBone >();
                rCollar_x_n025 = GetMorph( "RCollarX-025" );
                rCollar_x_p015 = GetMorph( "RCollarX+015" );
                rCollar_y_n017 = GetMorph( "RCollarY-017" );
                rCollar_y_p026 = GetMorph( "RCollarY+026" );
                rCollar_z_n050 = GetMorph( "RCollarZ-050" );
                rCollar_z_p015 = GetMorph( "RCollarZ+015" );
                drivers.Add( () => {
                    var reference = chest.rotation * rCollarBone.startingLocalRotation;

                    Vector3 rUp = (reference * Vector3.up).normalized,
                            rRight = (reference * Vector3.right).normalized,
                            rLeft = (reference * Vector3.left).normalized,
                            rForward = (reference * Vector3.forward).normalized;

                    var dazZRotation = AngleOffAroundAxis( rCollar.up, rUp, rForward, true );
                    rLog.AppendLine( $"{dazZRotation: 000.00;-000.00} rCollar Z" );
                    var unZ = Quaternion.AngleAxis( dazZRotation, rForward );

                    SetMorph( rCollar_z_n050, dazZRotation, 0, -50 );
                    SetMorph( rCollar_z_p015, dazZRotation, 0, 15 );

                    var dazYRotation = AngleOffAroundAxis( unZ * rCollar.right, rRight, rUp, true );
                    rLog.AppendLine( $"{dazYRotation: 000.00;-000.00} rCollar Y" );
                    var unY = Quaternion.AngleAxis( dazYRotation, rUp );

                    SetMorph( rCollar_y_n017, dazYRotation, 0, -17 );
                    SetMorph( rCollar_y_p026, dazYRotation, 0,  26 );

                    var dazXRotation = AngleOffAroundAxis( unY * unZ * rCollar.up, rUp, rLeft, true );
                    rLog.AppendLine( $"{dazXRotation: 000.00;-000.00} rCollar X" );

                    SetMorph( rCollar_x_n025, dazXRotation, 0, -25 );
                    SetMorph( rCollar_x_p015, dazXRotation, 0, 15 );
                            
                });
                #endregion

                #region Foot
                var lFootBone = lFoot.GetComponent< DAZBone >();
                lFoot_x_p065 = GetMorph( "LFootX+065" );
                lFoot_x_n040 = GetMorph( "LFootX-040" );
                drivers.Add( () => { 
                    var reference = lShin.rotation * lFootBone.startingLocalRotation;

                    Vector3 rLeft = (reference * Vector3.left).normalized,
                            rForward = (reference * Vector3.forward).normalized;

                    var dazXRotation = AngleOffAroundAxis( lFoot.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "lFoot X" );

                    SetMorph( lFoot_x_n040, dazXRotation, 0, -40 );
                    SetMorph( lFoot_x_p065, dazXRotation, 0, 65 );
                });

                var rFootBone = rFoot.GetComponent< DAZBone >();
                rFoot_x_p065 = GetMorph( "RFootX+065" );
                rFoot_x_n040 = GetMorph( "RFootX-040" );
                drivers.Add( () => {
                    var reference = rShin.rotation * rFootBone.startingLocalRotation;

                    Vector3 rLeft = (reference * Vector3.left).normalized,
                            rForward = (reference * Vector3.forward).normalized;

                    var dazXRotation = AngleOffAroundAxis( rFoot.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "rFoot X" );

                    SetMorph( rFoot_x_n040, dazXRotation, 0, -40 );
                    SetMorph( rFoot_x_p065, dazXRotation, 0, 65 );
                });
                #endregion

                #region Forearm
                var lForearmBone = lForearm.GetComponent< DAZBone >();
                lForearm_y_n100 = GetMorph( "LForearmY-100" );
                lForearm_y_n130 = GetMorph( "LForearmY-130" );
                drivers.Add( () => {
                    var reference = lShldr.rotation * lForearmBone.startingLocalRotation;

                    Vector3 rLeft = (reference * Vector3.left).normalized,
                            rUp = (reference * Vector3.up).normalized;

                    var dazYRotation = AngleOffAroundAxis( -lForearm.right, rLeft, rUp, true );
                    PrintRotation( dazYRotation, "lForeArm Y" );
                    var unY = Quaternion.AngleAxis( dazYRotation, rUp );

                    //SetMorph( lForearm_y_n060, dazYRotation, 0, -60 );
                    SetMorph( lForearm_y_n100, dazYRotation, 0, -100 );
                    SetMorph( lForearm_y_n130, dazYRotation, -100, -130 );
                } );

                var rForearmBone = rForearm.GetComponent< DAZBone >();
                rForearm_y_p100 = GetMorph( "RForearmY+100" );
                rForearm_y_p130 = GetMorph( "RForearmY+130" );
                drivers.Add( () => { 
                    var reference = rShldr.rotation * rForearmBone.startingLocalRotation;

                    Vector3 rRight = (reference * Vector3.right).normalized,
                            rUp = (reference * Vector3.up).normalized;

                    var dazYRotation = AngleOffAroundAxis( rForearm.right, rRight, rUp, true );
                    PrintRotation( dazYRotation, "rForeArm Y" );

                    //SetMorph( rForearm_y_p060, dazYRotation, 0, 60 );
                    SetMorph( rForearm_y_p100, dazYRotation, 0, 100 );
                    SetMorph( rForearm_y_p130, dazYRotation, 100, 130 );
                });
                #endregion

                #region Shin

                var lShinBone = lShin.GetComponent< DAZBone >();
                lShin_x_p085 = GetMorph( "LShinX+085" );
                lShin_x_p140 = GetMorph( "LShinX+140" );
                drivers.Add( () => {
                    var reference = lThigh.rotation * lShinBone.startingLocalRotation;

                    Vector3 rLeft    = (reference * Vector3.left).normalized,
                            rForward = (reference * Vector3.forward).normalized;

                    var dazXRotation = AngleOffAroundAxis( lShin.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "lShin X" );

                    SetMorph( lShin_x_p085, dazXRotation, 0, 85 );
                    SetMorph( lShin_x_p140, dazXRotation, 85, 140 );
                });

                var rShinBone = rShin.GetComponent< DAZBone >();
                rShin_x_p085 = GetMorph( "RShinX+085" );
                rShin_x_p140 = GetMorph( "RShinX+140" );
                drivers.Add( () => { 
                    var reference = rThigh.rotation * rShinBone.startingLocalRotation;
                    
                    Vector3 rLeft = (reference * Vector3.left).normalized,
                            rForward = (reference * Vector3.forward).normalized;

                    var dazXRotation = AngleOffAroundAxis( rShin.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "rShin X" );

                    SetMorph( rShin_x_p085, dazXRotation, 0, 85 );
                    SetMorph( rShin_x_p140, dazXRotation, 85, 140 );
                });
                #endregion

                #region Shoulder
                var lShldrBone = lShldr.GetComponent< DAZBone >();
                lShldr_z_p035 = GetMorph( "LShldrZ+035" );
                lShldr_z_n060 = GetMorph( "LShldrZ-060" );
                lShldr_z_n075 = GetMorph( "LShldrZ-075" );
                lShldr_y_n095 = GetMorph( "LShldrY-095" );
                lShldr_y_p040 = GetMorph( "LShldrY+040" );
                drivers.Add( () => {
                    // Get the reference position of the bone.
                    var reference = lCollar.rotation * lShldrBone.startingLocalRotation;

                    Vector3 rUp = (reference * Vector3.up).normalized,
                            rLeft = (reference * Vector3.left).normalized,
                            rForward = (reference * Vector3.forward).normalized;

                    var dazYRotation = AngleOffAroundAxis( -lShldr.right, rLeft, rUp, true );
                    PrintRotation( dazYRotation, "lShldr Y" );
                    var unY = Quaternion.AngleAxis( dazYRotation, rUp );

                    SetMorph( lShldr_y_n095, dazYRotation, 0, -95 );
                    SetMorph( lShldr_y_p040, dazYRotation, 0, 40 );

                    var dazZRotation = AngleOffAroundAxis( unY * -lShldr.right, rLeft, rForward, true );
                    PrintRotation( dazZRotation, "lShldr Z" );
                    var unZ = Quaternion.AngleAxis( dazZRotation, rForward );

                    SetMorph( lShldr_z_p035, dazZRotation, 0, 35 );
                    SetMorph( lShldr_z_n060, dazZRotation, 0, -60 );
                    SetMorph( lShldr_z_n075, dazZRotation, -60, -75 );

                    var dazXRotation = AngleOffAroundAxis( unZ * unY * lShldr.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "lShldr X" );
                });

                var rShldrBone = rShldr.GetComponent< DAZBone >();
                rShldr_z_n035 = GetMorph( "RShldrZ-035" );
                rShldr_z_p060 = GetMorph( "RShldrZ+060" );
                rShldr_z_p075 = GetMorph( "RShldrZ+075" );
                rShldr_y_n040 = GetMorph( "RShldrY-040" );
                rShldr_y_p095 = GetMorph( "RShldrY+095" );
                drivers.Add( () => { 
                    var reference = rCollar.rotation * rShldrBone.startingLocalRotation;
                    
                    Vector3 rUp = (reference * Vector3.up).normalized,
                            rRight = (reference * Vector3.right).normalized,
                            rForward = (reference * Vector3.forward).normalized;

                    var dazYRotation = AngleOffAroundAxis( rShldr.right, rRight, rUp, true );
                    PrintRotation( dazYRotation, "rShldr Y" );
                    var unY = Quaternion.AngleAxis( dazYRotation, rUp );

                    SetMorph( rShldr_y_n040, dazYRotation, 0, -40 );
                    SetMorph( rShldr_y_p095, dazYRotation, 0, 95 );

                    var dazZRotation = AngleOffAroundAxis( unY * rShldr.right, rRight, rForward, true );
                    PrintRotation( dazZRotation, "rShldr Z" );
                    var unZ = Quaternion.AngleAxis( dazZRotation, rForward );

                    SetMorph( rShldr_z_n035, dazZRotation, 0, -35 );
                    SetMorph( rShldr_z_p060, dazZRotation, 0, 60 );
                    SetMorph( rShldr_z_p075, dazZRotation, 60, 75 );

                    var dazXRotation = AngleOffAroundAxis( unZ * unY * rShldr.forward, rForward, rRight, true );
                    PrintRotation( dazXRotation, "rShldr X" );
                });
                #endregion

                #region Thigh
                lThigh_x_p035 = GetMorph( "LThighX+035" );
                lThigh_x_n055 = GetMorph( "LThighX-055" );
                lThigh_x_n115 = GetMorph( "LThighX-115" );
                lThigh_y_p075 = GetMorph( "LThighY+075" );
                lThigh_y_n075 = GetMorph( "LThighY-075" );
                lThigh_z_p085 = GetMorph( "LThighZ+085" );
                lThigh_z_n015 = GetMorph( "LThighZ-015" );

                rThigh_x_p035 = GetMorph( "RThighX+035" );
                rThigh_x_n055 = GetMorph( "RThighX-055" );
                rThigh_x_n115 = GetMorph( "RThighX-115" );
                rThigh_y_n075 = GetMorph( "RThighY-075" );
                rThigh_y_p075 = GetMorph( "RThighY+075" );
                rThigh_z_n085 = GetMorph( "RThighZ-085" );
                rThigh_z_p015 = GetMorph( "RThighZ+015" );

                cThighsZp180  = GetMorph( "CThighsZ+180" );
                cThighsZn030  = GetMorph( "CThighsZ-030" );
                cThighsXn115  = GetMorph( "CThighsX-115" );
                var lThighBone = lThigh.GetComponent< DAZBone >();
                var rThighBone = lThigh.GetComponent< DAZBone >();
                drivers.Add( () => { 
                    var lReference = pelvis.rotation * lThighBone.startingLocalRotation;
                    var rReference = pelvis.rotation * rThighBone.startingLocalRotation;

                    Vector3 lDown = (lReference * Vector3.down).normalized,
                            lLeft = (lReference * Vector3.left).normalized,
                            lForward = (lReference * Vector3.forward).normalized,
                            rDown = (rReference * Vector3.down).normalized,
                            rLeft = (rReference * Vector3.left).normalized,
                            rForward = (rReference * Vector3.forward).normalized;

                    var lThighXRotation = AngleOffAroundAxis( -lThigh.up, lDown, lLeft, true );
                    var rThighXRotation = AngleOffAroundAxis( -rThigh.up, rDown, rLeft, true );
                    PrintRotation( lThighXRotation, "lThigh X" );
                    PrintRotation( rThighXRotation, "rThigh X" );
                    
                    SetMorph( lThigh_x_p035, lThighXRotation, 0, 35 );
                    SetMorph( lThigh_x_n055, lThighXRotation, 0, -55 );
                    SetMorph( lThigh_x_n115, lThighXRotation, -55, -115 );
                    SetMorph( rThigh_x_p035, rThighXRotation, 0, 35 );
                    SetMorph( rThigh_x_n055, rThighXRotation, 0, -55 );
                    SetMorph( rThigh_x_n115, rThighXRotation, -55, -115 );

                    var thighsXAvg = (lThighXRotation + rThighXRotation) / 2;
                    PrintRotation( thighsXAvg, "CThighs X" );

                    SetMorph( cThighsXn115, thighsXAvg, -55, -115 );

                    var lUnX = Quaternion.AngleAxis( lThighXRotation, lLeft );
                    var rUnX = Quaternion.AngleAxis( rThighXRotation, rLeft );

                    var lThighZRotation = AngleOffAroundAxis( lUnX * -lThigh.up, lDown, lForward, true );
                    var rThighZRotation = AngleOffAroundAxis( rUnX * -rThigh.up, rDown, rForward, true );
                    PrintRotation( lThighZRotation, "lThigh Z" );
                    PrintRotation( rThighZRotation, "rThigh Z" );
                    
                    SetMorph( lThigh_z_n015, lThighZRotation, 0, -15 );
                    SetMorph( lThigh_z_p085, lThighZRotation, 0, 85 );
                    SetMorph( rThigh_z_n085, rThighZRotation, 0, -85 );
                    SetMorph( rThigh_z_p015, rThighZRotation, 0, 15 );

                    var thighZSeparation = lThighZRotation - rThighZRotation;

                    PrintRotation( thighZSeparation, "thigh Z separation");
                    SetMorph( cThighsZp180, thighZSeparation, 0, 180 );
                    SetMorph( cThighsZn030, thighZSeparation, 0, -30 );

                    var lUnZ = Quaternion.AngleAxis( lThighZRotation, lForward );
                    var rUnZ = Quaternion.AngleAxis( rThighZRotation, rForward );

                    // Trial and error got me to this odd un-rotation that seems to produce the most
                    // accurate Y rotation.
                    var lThighYRotation = AngleOffAroundAxis( lUnX * lUnZ * lThigh.forward, lForward, lThigh.up, true );
                    var rThighYRotation = AngleOffAroundAxis( rUnX * rUnZ * rThigh.forward, rForward, rThigh.up, true );
                    PrintRotation( lThighYRotation, "lThigh Y" );
                    PrintRotation( rThighYRotation, "rThigh Y" );
                    
                    SetMorph( lThigh_y_n075, lThighYRotation, 0, -75 );
                    SetMorph( lThigh_y_p075, lThighYRotation, 0, 75 );
                    SetMorph( rThigh_y_n075, rThighYRotation, 0, -75 );
                    SetMorph( rThigh_y_p075, rThighYRotation, 0, 75 );

                });
                #endregion

                #region Trunk
                var pelvisBone = pelvis.GetComponent< DAZBone >();
                pelvis_x_p030 = GetMorph( "TPelvisX+030" );
                pelvis_x_n015 = GetMorph( "TPelvisX-015" );
                drivers.Add( () => { 
                    var reference = hip.rotation * pelvisBone.startingLocalRotation;

                    Vector3 rForward = (reference * Vector3.forward).normalized,
                            rLeft = (reference * Vector3.left).normalized;

                    var dazXRotation = AngleOffAroundAxis( pelvis.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "pelvis X" );
                    var unX = Quaternion.AngleAxis( dazXRotation, rLeft );

                    SetMorph( pelvis_x_n015, dazXRotation, 0, -15 );
                    SetMorph( pelvis_x_p030, dazXRotation, 0, 30 );
                });

                var abdomenBone = abdomen.GetComponent< DAZBone >();
                abdomen_x_n020 = GetMorph( "TAbdomenX-020" );
                abdomen_x_p030 = GetMorph( "TAbdomenX+030" );
                drivers.Add( () => { 
                    var reference = hip.rotation * abdomenBone.startingLocalRotation;

                    Vector3 rForward = (reference * Vector3.forward).normalized,
                            rLeft = (reference * Vector3.left).normalized;

                    var dazXRotation = AngleOffAroundAxis( abdomen.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "abdomen X" );

                    SetMorph( abdomen_x_p030, dazXRotation, 0, 30 );
                    SetMorph( abdomen_x_n020, dazXRotation, 0, -20 );
                });

                var abdomen2Bone = abdomen2.GetComponent< DAZBone >();
                abdomen2_x_n020 = GetMorph( "TAbdomen2X-020" );
                abdomen2_x_p030 = GetMorph( "TAbdomen2X+030" );
                drivers.Add( () => { 
                    var reference = abdomen.rotation * abdomen2Bone.startingLocalRotation;

                    Vector3 rForward = (reference * Vector3.forward).normalized,
                            rLeft = (reference * Vector3.left).normalized;

                    var dazXRotation = AngleOffAroundAxis( abdomen2.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "abdomen2 X" );

                    SetMorph( abdomen2_x_n020, dazXRotation, 0, -20 );
                    SetMorph( abdomen2_x_p030, dazXRotation, 0, 30 );
                });

                var chestBone = chest.GetComponent< DAZBone >();
                chest_x_p020 = GetMorph( "TChestX+020" );
                drivers.Add( () => { 
                    var reference = abdomen2.rotation * chestBone.startingLocalRotation;

                    Vector3 rForward = (reference * Vector3.forward).normalized,
                            rLeft = (reference * Vector3.left).normalized;

                    var dazXRotation = AngleOffAroundAxis( chest.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "chest X" );

                    SetMorph( chest_x_p020, dazXRotation, 0, 20 );
                });

                var neckBone = neck.GetComponent< DAZBone >();
                neck_x_n030 = GetMorph( "TNeckX-030" );
                drivers.Add( () => { 
                    var reference = chest.rotation * neckBone.startingLocalRotation;

                    Vector3 rForward = (reference * Vector3.forward).normalized,
                            rLeft = (reference * Vector3.left).normalized;

                    var dazXRotation = AngleOffAroundAxis( neck.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "neck X" );
                    
                    SetMorph( neck_x_n030, dazXRotation, 0, -30 );
                });

                var headBone = head.GetComponent< DAZBone >();
                head_x_n045 = GetMorph( "THeadX-045" );
                drivers.Add( () => { 
                    var reference = neck.rotation * headBone.startingLocalRotation;

                    Vector3 rForward = (reference * Vector3.forward).normalized,
                            rLeft = (reference * Vector3.left).normalized;

                    var dazXRotation = AngleOffAroundAxis( head.forward, rForward, rLeft, true );
                    PrintRotation( dazXRotation, "head X" );

                    SetMorph( head_x_n045, dazXRotation, 0, -45 );
                });
                #endregion

                #if DEBUG_LINE
                { 
                    line = gameObject.GetComponent< LineRenderer >();
                    if( line == null ) { 
                        line = gameObject.AddComponent< LineRenderer >();
                    }

                    line.positionCount = 5;
                    foreach( Material mat in line.materials )
                    { 
                        line.startColor = Color.blue;
                        line.endColor = Color.blue;
                        line.startWidth = 0.04f;
                        line.endWidth = 0.00f;
                    }
                }
                #endif

                #if DEBUG
                { 
                    var NotFoundArray = NotFound.ToArray();
                    drivers.Add( () => { 
                        if( NotFoundArray.Length > 0 )
                        {
                            mLog.AppendLine( $"Missing morphs:\n  {string.Join( ",\n  ", NotFoundArray )}" );
                        }
                    });
                }
                #endif

                locked.val = false;
            }
            catch ( Exception ex )
            {
                SuperController.LogError( $"JointCorrect did not Init.\n{ex}" );
                locked.val = true;
            }
        }

        public override void InitUI()
        {
            base.InitUI();
            try
            { 
                #region Variants UI
                variantUIs = new Dictionary<string, JSONStorableBool>();
                foreach( var variantName in allVariants )
                {
                    var viewName = $"Variant: {variantName}";
                    if( variantName == "" )
                        viewName = "JointCorrect: Base";
                    else if( variantName == "gens" )
                        viewName = "JointCorrect: Genitalia";

                    var sw = new JSONStorableBool( viewName, enabledVariants.Contains( variantName ), v => ToggleVariant( v, variantName ) );
                    sw.altName = variantName;
                    variantUIs[ variantName ] = sw;
                    CreateToggle( sw, false );
                }
                #endregion
            }
            catch( Exception ex )
            {
                SuperController.LogError( $"JointCorrect did not InitUI.\n{ex}" );
                locked.val = true;
            }
        }

        public void Update()
        {
            if( locked.val )
                return;
            try
            {
                { 
                    mLog.Remove( 0, mLog.Length );
                    rLog.Remove( 0, rLog.Length );
                }
                
                foreach( Action driver in drivers )
                {
                    driver.Invoke();
                }

                #if DEBUG
                { 
                    morphOut.valNoCallback = mLog.ToString();
                    jointOut.valNoCallback = rLog.ToString();
                }
                #endif
            }
            catch( Exception ex )
            {
                locked.val = true;
                SuperController.LogError( "Exception calculating control morphs: " + ex );
            }
        }

        public void OnDisable()
        {
            foreach( MorphVariant m in allMorphs )
               m.morph.morphValue = 0;
        }

        private void VariantsChanged( HashSet< string > enabledVariants )
        {
            foreach( string variant in allVariants )
            {
                JSONStorableBool vbool = null;
                if( variantUIs.TryGetValue( variant, out vbool ) )
                { 
                    vbool.val = enabledVariants.Contains( variant );
                }
            }
        }

        private void ToggleVariant( bool enabled, string variantName )
        {
            if( enabled ) 
                enabledVariants.Add( variantName );
            else
                enabledVariants.Remove( variantName );
        }

        private void DrawAxes( Quaternion rotation, Vector3 position )
        {
            if( ! debugLine )
                return;

            var length = 0.25f;
            
            line.positionCount = 5;
            
            line.startColor = Color.green;
            line.endColor = Color.blue;

            line.SetPosition( 0, position + ((rotation * Vector3.up).normalized * length) );
            line.SetPosition( 1, position );
            line.SetPosition( 2, position + ((rotation * Vector3.right).normalized * length) );
            line.SetPosition( 3, position );
            line.SetPosition( 4, position + ((rotation * Vector3.forward).normalized * length) );
        }

        private void SetMorph( MorphVariant[] morphs, float value, float starting, float ending )
        {
            int mask = builtin.val ? BUILTIN : EXTERNAL;
            value = NormalizeFloat( value, starting, ending );
            foreach( MorphVariant m in morphs )
            {
                if( enabledVariants.Contains( m.variant ) && (m.mask & mask) > 0 )
                { 
                    #if DEBUG
                        mLog.AppendLine( $"{m.morph.displayName} <- {value:##0.00}" );
                    #endif
                    m.morph.morphValue = value;
                }
                else
                {
                    m.morph.morphValue = 0;
                }
            }
        }

        private MorphVariant[] GetMorph( string name )
        {
            var mc = (person.GetStorableByID( "geometry" ) as DAZCharacterSelector).morphsControlUI;
            var morphsList = new List< MorphVariant >();
            foreach( DAZMorph morph in mc.GetMorphs() )
            {
                var names = morph.displayName.Split( '.' );
                if( names[0].Equals( name ) )
                {
                    string variant = "";

                    if( names.Length > 1 )
                    {
                        variant = names[1];

                        if( ! allVariants.Contains( variant ) )
                        { 
                            //SuperController.LogMessage( $"Found JCM variant {variant}" );
                            allVariants.Add( variant );
                        }
                    }

                    if( morph.isLatestVersion )
                    { 
                        if( variant == "" || variant == "gens" )
                        {
                            if( morph.isLatestVersion )
                                morphsList.Add( new MorphVariant( variant, morph, morph.isInPackage ? BUILTIN : EXTERNAL ) );
                        }
                        else
                        {
                            morphsList.Add( new MorphVariant( variant, morph, VARIANT ) );
                        }
                    }
                }
            }

            allMorphs.AddRange( morphsList );
            
            if( morphsList.Count == 0 )
            {
                NotFound.Add( name );
            }

            return morphsList.ToArray();
        }

        private Transform GetJoint( string name )
        {
            return person.GetStorableByID( name ).transform;
        }

        private void PrintRotation( double value, string name )
        {
            #if DEBUG
                rLog.AppendLine( $"{value: 000.00;-000.00} {name}" );
            #endif
        }

        private static float NormalizeFloat( float value, float start, float end )
        {
            return Mathf.Clamp( (value - start) / (end - start), 0, 1 );
        }

        static public float CalcAxisRotation( Vector3 question, Vector3 baseline, Vector3 axis, bool anti = false )
        {
            Vector3 right;
            if( anti )
            {
                right = Vector3.Cross( baseline, axis );
                baseline = Vector3.Cross( axis, right );
            }
            else
            {
                right = Vector3.Cross( axis, baseline );
                baseline = Vector3.Cross( right, axis );
            }

            return Mathf.Atan2(Vector3.Dot( question, right ), Vector3.Dot( question, baseline ) ) * 180 / Mathf.PI;
        }

        static public float CalcSignedTransformAngle( Transform from, Transform thru, Transform to, Vector3 approxNormal, bool clockwise = false )
        {
            Vector3 baseline = (thru.position - from.position).normalized;
            Vector3 question = (thru.position - to.position).normalized;
            Vector3 tangent  = Vector3.Cross( baseline, question );

            var transformAngle = Vector3.Angle( baseline, question );


            return 0;
        }

        static public float AngleOffAroundAxis( Vector3 v, Vector3 forward, Vector3 axis, bool clockwise = false )
        {
            Vector3 right;
            if( clockwise )
            {
                right = Vector3.Cross( forward, axis );
                forward = Vector3.Cross( axis, right );
            }
            else
            {
                right = Vector3.Cross( axis, forward );
                forward = Vector3.Cross( right, axis );
            }

            return Mathf.Atan2(Vector3.Dot( v, right ), Vector3.Dot( v, forward ) ) * 180 / Mathf.PI;
        }

        internal const int BUILTIN = 1;
        internal const int EXTERNAL = 2;
        internal const int VARIANT = 3;

        private class MorphVariant
        {
            public int mask;
            public string   variant;
            public DAZMorph morph;

            public MorphVariant( string variant, DAZMorph morph, int mask )
            {
                this.variant = variant;
                this.morph   = morph;
                this.mask = mask;
            }

            public bool IsSame( string[] names )
            {
                if( names[0] == morph.displayName )
                {
                    string var = names.Length > 1 ? names[1] : "";
                    return variant == var;
                }

                return false;
            }
        }

        public class StorableStringSet : JSONStorableString
        {
            private HashSet< string > underlying;
            public delegate void SetStringSetCallback( HashSet< string > value );
            
            public new SetStringSetCallback setCallbackFunction { get; set; }

            public StorableStringSet( string paramName, SetStringSetCallback callback ) : base( paramName, "<proxy>" )
            {
                underlying = new HashSet< string >();
                setCallbackFunction = callback;
            }

            public void Add( string st )
            {
                if( underlying.Add( st ) )
                    setCallbackFunction( underlying );
            }


            public void Remove( string st )
            {
                if( underlying.Remove( st ) )
                    setCallbackFunction( underlying );
            }

            public bool Contains( string st )
            {
                return underlying.Contains( st );
            }

            public override void RestoreFromJSON( JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true )
            {
                underlying.Clear();
                foreach( JSONNode node in jc[name].AsArray )
                {
                    string s = node.Value;
                    underlying.Add( s );
                }

                if( underlying.Count == 0 )
                {
                    underlying.Add( "" );
                    underlying.Add( "gens" );
                }

                setCallbackFunction( underlying );
            }

            public override bool StoreJSON( JSONClass jc, bool includePhysical = true, bool includeAppearance = true, bool forceStore = false )
            {
                var ar = new JSONArray();
                foreach( string s in underlying )
                {
                    ar.Add( s );
                }
                jc[name] = ar;

                return true;
            }
        }
    }
}
